@@ -37,17 +37,18 @@ struct logInfoHead logs;
 #define REALLOC_STEP            10
 #define GLOB_STR_REALLOC_STEP   0x100
 
-#if defined(SunOS)
-#include <limits.h>
-#if !defined(isblank)
-#define isblank(c) ( (c) == ' ' || (c) == '\t' ) ? 1 : 0
-#endif
+#if defined(SunOS) && !defined(isblank)
+#define isblank(c) ( ( (c) == ' ' || (c) == '\t' ) ? 1 : 0 )
 #endif
 
 #ifdef __hpux
 #include "asprintf.c"
 #endif
 
+#if !defined(HAVE_SECURE_GETENV)
+#define secure_getenv getenv
+#endif
+
 #if !defined(HAVE_ASPRINTF) && !defined(_FORTIFY_SOURCE)
 #include <stdarg.h>
 
@@ -70,7 +71,7 @@ int asprintf(char **string_ptr, const char *format, ...)
          * Strictly speaking, GNU asprintf doesn't do this,
          * but the caller isn't checking the return value.
          */
-        fprintf(stderr, "failed to allocate memory\\n");
+        message_OOM();
         exit(1);
     }
     rv = vsnprintf(str, size, format, arg);
@@ -88,9 +89,6 @@ char *strndup(const char *s, size_t n)
     size_t nAvail;
     char *p;
 
-    if(!s)
-        return NULL;
-
     /* min() */
     nAvail = strlen(s) + 1;
     if ( (n + 1) < nAvail)
@@ -114,10 +112,11 @@ static const struct compress_cmd_item compress_cmd_list[] = {
     {"gzip", ".gz"},
     {"bzip2", ".bz2"},
     {"xz", ".xz"},
+    {"zstd", ".zst"},
     {"compress", ".Z"},
     {"zip", "zip"},
 };
-static const int compress_cmd_list_size = sizeof(compress_cmd_list)
+static const unsigned compress_cmd_list_size = sizeof(compress_cmd_list)
     / sizeof(compress_cmd_list[0]);
 
 enum {
@@ -131,6 +130,7 @@ enum {
 
 static const char *defTabooExts[] = {
     ",v",
+    ".bak",
     ".cfsaved",
     ".disabled",
     ".dpkg-bak",
@@ -149,11 +149,11 @@ static const char *defTabooExts[] = {
     ".ucf-old",
     "~"
 };
-static const int defTabooCount = sizeof(defTabooExts) / sizeof(char *);
+static const unsigned defTabooCount = sizeof(defTabooExts) / sizeof(char *);
 
 /* I shouldn't use globals here :-( */
 static char **tabooPatterns = NULL;
-static int tabooCount = 0;
+static unsigned tabooCount = 0;
 static int glob_errno = 0;
 
 static int readConfigFile(const char *configFile, struct logInfo *defConfig);
@@ -173,7 +173,11 @@ static char *isolateLine(char **strt, char **buf, size_t length) {
     tmp = endtag - 1;
     while (isspace((unsigned char)*endtag))
         endtag--;
-    key = strndup(start, endtag - start + 1);
+    key = strndup(start, (size_t)(endtag - start + 1));
+    if (key == NULL) {
+        message_OOM();
+        return NULL;
+    }
     *strt = tmp;
     return key;
 }
@@ -214,7 +218,11 @@ static char *isolateWord(char **strt, char **buf, size_t length) {
         endtag++;}
     if (max < endtag)
         return NULL;
-    key = strndup(start, endtag - start);
+    key = strndup(start, (size_t)(endtag - start));
+    if (key == NULL) {
+        message_OOM();
+        return NULL;
+    }
     *strt = endtag;
     return key;
 }
@@ -226,10 +234,10 @@ static char *readPath(const char *configFile, int lineNum, const char *key,
     if (path != NULL) {
         wchar_t pwc;
         size_t len;
-        char *chptr = path;
+        const char *chptr = path;
 
         while (*chptr && (len = mbrtowc(&pwc, chptr, strlen(chptr), NULL)) != 0) {
-            if (len == (size_t)(-1) || len == (size_t)(-2) || !iswprint(pwc) || iswblank(pwc)) {
+            if (len == (size_t)(-1) || len == (size_t)(-2) || !iswprint((wint_t)pwc) || iswblank((wint_t)pwc)) {
                 message(MESS_ERROR, "%s:%d bad %s path %s\n",
                         configFile, lineNum, key, path);
                 free(path);
@@ -244,37 +252,67 @@ static char *readPath(const char *configFile, int lineNum, const char *key,
 /* set *pUid to UID of the given user, return non-zero on failure */
 static int resolveUid(const char *userName, uid_t *pUid)
 {
-    struct passwd *pw;
+    const struct passwd *pw;
+    char *endptr;
+    unsigned long int parsed_uid;
+
 #ifdef __CYGWIN__
     if (strcmp(userName, "root") == 0) {
         *pUid = 0;
         return 0;
     }
 #endif
+
     pw = getpwnam(userName);
-    if (!pw)
-        return -1;
-    *pUid = pw->pw_uid;
-    endpwent();
-    return 0;
+    if (pw) {
+        *pUid = pw->pw_uid;
+        return 0;
+    }
+
+    parsed_uid = strtoul(userName, &endptr, 10);
+    if (userName[0] != '\0' &&
+        *endptr == '\0' &&
+        parsed_uid < INT_MAX && /* parsed_uid != ULONG_MAX && */
+        getpwuid((uid_t)parsed_uid) != NULL) {
+
+        *pUid = (uid_t)parsed_uid;
+        return 0;
+    }
+
+    return -1;
 }
 
 /* set *pGid to GID of the given group, return non-zero on failure */
 static int resolveGid(const char *groupName, gid_t *pGid)
 {
-    struct group *gr;
+    const struct group *gr;
+    char *endptr;
+    unsigned long int parsed_gid;
+
 #ifdef __CYGWIN__
     if (strcmp(groupName, "root") == 0) {
         *pGid = 0;
         return 0;
     }
 #endif
+
     gr = getgrnam(groupName);
-    if (!gr)
-        return -1;
-    *pGid = gr->gr_gid;
-    endgrent();
-    return 0;
+    if (gr) {
+        *pGid = gr->gr_gid;
+        return 0;
+    }
+
+    parsed_gid = strtoul(groupName, &endptr, 10);
+    if (groupName[0] != '\0' &&
+        *endptr == '\0' &&
+        parsed_gid < INT_MAX && /* parsed_gid != ULONG_MAX && */
+        getgrgid((gid_t)parsed_gid) != NULL) {
+
+        *pGid = (gid_t)parsed_gid;
+        return 0;
+    }
+
+    return -1;
 }
 
 static int readModeUidGid(const char *configFile, int lineNum, char *key,
@@ -282,15 +320,18 @@ static int readModeUidGid(const char *configFile, int lineNum, char *key,
                           gid_t *pGid)
 {
     char u[200], g[200];
-    unsigned int m = 0;
+    mode_t m = 0;
     char tmp;
     int rc;
 
     if (!strcmp("su", directive))
         /* do not read <mode> for the 'su' directive */
         rc = 0;
-    else
-        rc = sscanf(key, "%o %199s %199s%c", &m, u, g, &tmp);
+    else {
+        unsigned short int parsed_mode;
+        rc = sscanf(key, "%ho %199s %199s%c", &parsed_mode, u, g, &tmp);
+        m = parsed_mode;
+    }
 
     /* We support 'key <owner> <group> notation now */
     if (rc == 0) {
@@ -338,7 +379,7 @@ static char *readAddress(const char *configFile, int lineNum, const char *key,
 
     if (address != NULL) {
         /* validate the address */
-        char *chptr = address;
+        const char *chptr = address;
         while (isprint((unsigned char) *chptr) && *chptr != ' ') {
             chptr++;
         }
@@ -394,7 +435,7 @@ static int mkpath(const char *path, mode_t mode, uid_t uid, gid_t gid) {
     char *copypath = strdup(path);
 
     if (!copypath) {
-        message(MESS_ERROR, "can not allocate memory\n");
+        message_OOM();
         return 1;
     }
 
@@ -417,7 +458,7 @@ static int mkpath(const char *path, mode_t mode, uid_t uid, gid_t gid) {
 
 static int checkFile(const char *fname)
 {
-    int i;
+    unsigned i;
 
     /* Check if fname is '.' or '..'; if so, return false */
     if (fname[0] == '.' && (!fname[1] || (fname[1] == '.' && !fname[2])))
@@ -444,9 +485,9 @@ static int compar(const void *p, const void *q)
 }
 
 /* Free memory blocks pointed to by pointers in a 2d array and the array itself */
-static void free_2d_array(char **array, int lines_count)
+static void free_2d_array(char **array, unsigned lines_count)
 {
-    int i;
+    unsigned i;
     for (i = 0; i < lines_count; ++i)
         free(array[i]);
     free(array);
@@ -454,10 +495,10 @@ static void free_2d_array(char **array, int lines_count)
 
 #define MEMBER_COPY(dest, src) \
     do { \
-        if (src && rv == 0) { \
+        if ((src) && rv == 0) { \
             (dest) = strdup(src); \
             if ((dest) == NULL) { \
-                message(MESS_ERROR, "can not allocate memory\n"); \
+                message_OOM(); \
                 rv = 1; \
             } \
         } else { \
@@ -504,7 +545,7 @@ static int copyLogInfo(struct logInfo *to, const struct logInfo *from)
         poptDupArgv(from->compress_options_count, from->compress_options_list,
                     &to->compress_options_count,  &to->compress_options_list);
         if (to->compress_options_list == NULL) {
-            message(MESS_ERROR, "can not allocate memory\n");
+            message_OOM();
             rv = 1;
         }
     }
@@ -542,7 +583,7 @@ static struct logInfo *newLogInfo(const struct logInfo *template)
 
     new = malloc(sizeof(*new));
     if (new == NULL) {
-        message(MESS_ERROR, "can not allocate memory\n");
+        message_OOM();
         return NULL;
     }
 
@@ -614,9 +655,10 @@ static int readConfigPath(const char *path, struct logInfo *defConfig)
     }
 
     if (S_ISDIR(sb.st_mode)) {
-        char **namelist, **p;
+        char **namelist = NULL;
         struct dirent *dp;
-        int files_count, here, i;
+        int here;
+        unsigned files_count = 0, i;
         DIR *dirp;
 
         if ((here = open(".", O_RDONLY)) == -1) {
@@ -631,13 +673,11 @@ static int readConfigPath(const char *path, struct logInfo *defConfig)
             close(here);
             return 1;
         }
-        files_count = 0;
-        namelist = NULL;
         while ((dp = readdir(dirp)) != NULL) {
             if (checkFile(dp->d_name)) {
                 /* Realloc memory for namelist array if necessary */
                 if (files_count % REALLOC_STEP == 0) {
-                    p = (char **) realloc(namelist,
+                    char **p = (char **) realloc(namelist,
                             (files_count +
                              REALLOC_STEP) * sizeof(char *));
                     if (p) {
@@ -648,8 +688,7 @@ static int readConfigPath(const char *path, struct logInfo *defConfig)
                         free_2d_array(namelist, files_count);
                         closedir(dirp);
                         close(here);
-                        message(MESS_ERROR, "cannot realloc: %s\n",
-                                strerror(errno));
+                        message_OOM();
                         return 1;
                     }
                 }
@@ -661,8 +700,7 @@ static int readConfigPath(const char *path, struct logInfo *defConfig)
                     free_2d_array(namelist, files_count);
                     closedir(dirp);
                     close(here);
-                    message(MESS_ERROR, "cannot realloc: %s\n",
-                            strerror(errno));
+                    message_OOM();
                     return 1;
                 }
             }
@@ -704,7 +742,7 @@ static int readConfigPath(const char *path, struct logInfo *defConfig)
         }
 
         if (fchdir(here) < 0) {
-            message(MESS_ERROR, "could not change directory to '.'");
+            message(MESS_ERROR, "could not change directory to '.': %s\n", strerror(errno));
         }
         close(here);
         free_2d_array(namelist, files_count);
@@ -727,7 +765,8 @@ static int readConfigPath(const char *path, struct logInfo *defConfig)
 
 int readAllConfigPaths(const char **paths)
 {
-    int i, result = 0;
+    int result = 0;
+    unsigned i;
     const char **file;
     struct logInfo defConfig = {
         .pattern = NULL,
@@ -770,7 +809,7 @@ int readAllConfigPaths(const char **paths)
 
     tabooPatterns = malloc(sizeof(*tabooPatterns) * defTabooCount);
     if (tabooPatterns == NULL) {
-        message(MESS_ERROR, "can not allocate memory\n");
+        message_OOM();
         return 1;
     }
 
@@ -788,7 +827,7 @@ int readAllConfigPaths(const char **paths)
             tabooCount++;
         } else {
             free_2d_array(tabooPatterns, tabooCount);
-            message(MESS_ERROR, "cannot malloc: %s\n", strerror(errno));
+            message_OOM();
             return 1;
         }
     }
@@ -803,7 +842,7 @@ int readAllConfigPaths(const char **paths)
 }
 
 static char* parseGlobString(const char *configFile, int lineNum,
-                             const char *buf, off_t length, char **ppos)
+                             const char *buf, size_t length, char **ppos)
 {
     /* output buffer */
     char *globString = NULL;
@@ -816,7 +855,7 @@ static char* parseGlobString(const char *configFile, int lineNum,
     } state = PGS_INIT;
 
     /* move the cursor at caller's side while going through the input */
-    for (; (*ppos - buf < length) && **ppos; (*ppos)++) {
+    for (; ((size_t)(*ppos - buf) < length) && **ppos; (*ppos)++) {
         /* state transition (see above) */
         switch (state) {
             case PGS_INIT:
@@ -857,7 +896,7 @@ static char* parseGlobString(const char *configFile, int lineNum,
             globStringAlloc += GLOB_STR_REALLOC_STEP;
             ptr = realloc(globString, globStringAlloc);
             if (!ptr) {
-                message(MESS_ERROR, "can not realloc\n");
+                message_OOM();
                 free(globString);
                 return NULL;
             }
@@ -906,7 +945,7 @@ static int globerr(const char *pathname, int theerr)
 #define RAISE_ERROR() \
     if (newlog != defConfig) { \
         state = STATE_ERROR; \
-        continue; \
+        goto next_state; \
     } else { \
         goto error; \
     }
@@ -915,31 +954,21 @@ static int globerr(const char *pathname, int theerr)
 static int readConfigFile(const char *configFile, struct logInfo *defConfig)
 {
     int fd;
-    char *buf, *endtag, *key = NULL;
-    off_t length;
+    char *buf, *key = NULL;
+    size_t length;
     int lineNum = 1;
-    unsigned long long multiplier;
-    int i, k;
     char *scriptStart = NULL;
     char **scriptDest = NULL;
     struct logInfo *newlog = defConfig;
     char *start, *chptr;
-    struct passwd *pw = NULL;
-    int rc;
-    struct stat sb, sb2;
-    glob_t globResult;
-    const char **argv;
-    int argc, argNum;
-    int flags;
+    struct stat sb;
     int state = STATE_DEFAULT;
     int logerror = 0;
-    struct logInfo *log;
     /* to check if incompatible criteria are specified */
     int criterium_set = 0;
     static unsigned recursion_depth = 0U;
     char *globerr_msg = NULL;
     int in_config = 0;
-    int rv;
     struct flock fd_lock = {
         .l_start = 0,
         .l_len = 0,
@@ -953,19 +982,6 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
                 configFile, strerror(errno));
         return 1;
     }
-    if ((flags = fcntl(fd, F_GETFD)) == -1) {
-        message(MESS_ERROR, "Could not retrieve flags from file %s\n",
-                configFile);
-        close(fd);
-        return 1;
-    }
-    flags |= FD_CLOEXEC;
-    if (fcntl(fd, F_SETFD, flags) == -1) {
-        message(MESS_ERROR, "Could not set flags on file %s\n",
-                configFile);
-        close(fd);
-        return 1;
-    }
     /* We don't want anybody to change the file while we parse it,
      * let's try to lock it for reading. */
     if (fcntl(fd, F_SETLK, &fd_lock) == -1) {
@@ -986,8 +1002,9 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
         return 0;
     }
 
-    if (!(pw = getpwuid(getuid()))) {
-        message(MESS_ERROR, "Logrotate UID is not in passwd file.\n");
+    if (!getpwuid(getuid())) {
+        message(MESS_ERROR, "Cannot find logrotate UID (%d) in passwd file: %s\n",
+                getuid(), strerror(errno));
         close(fd);
         return 1;
     }
@@ -1007,17 +1024,7 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
             return 0;
         }
 
-        if ((pw = getpwuid(ROOT_UID)) == NULL) {
-            message(MESS_ERROR,
-                    "Ignoring %s because there's no password entry for the owner.\n",
-                    configFile);
-            close(fd);
-            return 0;
-        }
-
-        if (sb.st_uid != ROOT_UID && (pw == NULL ||
-                    sb.st_uid != pw->pw_uid ||
-                    pw->pw_uid != ROOT_UID)) {
+        if (sb.st_uid != ROOT_UID) {
             message(MESS_ERROR,
                     "Ignoring %s because the file owner is wrong (should be root or user with uid 0).\n",
                     configFile);
@@ -1026,7 +1033,7 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
         }
     }
 
-    length = sb.st_size;
+    length = (size_t)sb.st_size;
 
     if (length > 0xffffff) {
         message(MESS_ERROR, "file %s too large, probably not a config file.\n",
@@ -1045,10 +1052,10 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
     }
 
 #ifdef MAP_POPULATE
-    buf = mmap(NULL, (size_t) length, PROT_READ,
+    buf = mmap(NULL, length, PROT_READ,
             MAP_PRIVATE | MAP_POPULATE, fd, (off_t) 0);
 #else /* MAP_POPULATE */
-    buf = mmap(NULL, (size_t) length, PROT_READ,
+    buf = mmap(NULL, length, PROT_READ,
             MAP_PRIVATE, fd, (off_t) 0);
 #endif /* MAP_POPULATE */
 
@@ -1061,17 +1068,17 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
 
 #ifdef HAVE_MADVISE
 #ifdef MADV_DONTFORK
-    madvise(buf, (size_t)(length + 2),
+    madvise(buf, length + 2,
             MADV_SEQUENTIAL | MADV_WILLNEED | MADV_DONTFORK);
 #else /* MADV_DONTFORK */
-    madvise(buf, (size_t)(length + 2),
+    madvise(buf, length + 2,
             MADV_SEQUENTIAL | MADV_WILLNEED);
 #endif /* MADV_DONTFORK */
 #endif /* HAVE_MADVISE */
 
     message(MESS_DEBUG, "reading config file %s\n", configFile);
 
-    for (start = buf; start - buf < length; start++) {
+    for (start = buf; (size_t)(start - buf) < length; start++) {
         switch (state) {
             case STATE_DEFAULT:
                 if (isblank((unsigned char)*start))
@@ -1139,6 +1146,7 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
                     } else if (!strcmp(key, "maillast")) {
                         newlog->flags &= ~LOG_FLAG_MAILFIRST;
                     } else if (!strcmp(key, "su")) {
+                        int rv;
                         mode_t tmp_mode = NO_MODE;
                         free(key);
                         key = isolateLine(&start, &buf, length);
@@ -1169,6 +1177,8 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
 
                         newlog->flags |= LOG_FLAG_SU;
                     } else if (!strcmp(key, "create")) {
+                        int rv;
+
                         free(key);
                         key = isolateLine(&start, &buf, length);
                         if (key == NULL)
@@ -1183,6 +1193,8 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
 
                         newlog->flags |= LOG_FLAG_CREATE;
                     } else if (!strcmp(key, "createolddir")) {
+                        int rv;
+
                         free(key);
                         key = isolateLine(&start, &buf, length);
                         if (key == NULL)
@@ -1202,12 +1214,13 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
                         newlog->flags &= ~LOG_FLAG_CREATE;
                     } else if (!strcmp(key, "size") || !strcmp(key, "minsize") ||
                             !strcmp(key, "maxsize")) {
-                        unsigned long long size = 0;
                         char *opt = key;
 
                         key = isolateValue(configFile, lineNum, opt, &start, &buf, length);
                         if (key && key[0]) {
-                            int l = strlen(key) - 1;
+                            off_t size;
+                            unsigned long multiplier;
+                            const size_t l = strlen(key) - 1;
                             if (key[l] == 'k' || key[l] == 'K') {
                                 key[l] = '\0';
                                 multiplier = 1024;
@@ -1226,8 +1239,8 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
                                 multiplier = 1;
                             }
 
-                            size = multiplier * strtoull(key, &chptr, 0);
-                            if (*chptr) {
+                            size = (off_t) (multiplier * strtoull(key, &chptr, 0));
+                            if (*chptr || size < 0) {
                                 message(MESS_ERROR, "%s:%d bad size '%s'\n",
                                         configFile, lineNum, key);
                                 free(opt);
@@ -1253,7 +1266,7 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
                                            &start, &buf, length);
                         if (key == NULL)
                             continue;
-                        newlog->shred_cycles = strtoul(key, &chptr, 0);
+                        newlog->shred_cycles = (int)strtoul(key, &chptr, 0);
                         if (*chptr || newlog->shred_cycles < 0) {
                             message(MESS_ERROR, "%s:%d bad shred cycles '%s'\n",
                                     configFile, lineNum, key);
@@ -1294,7 +1307,7 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
                                            &buf, length);
                         if (key == NULL)
                             continue;
-                        newlog->rotateCount = strtol(key, &chptr, 0);
+                        newlog->rotateCount = (int)strtol(key, &chptr, 0);
                         if (*chptr || newlog->rotateCount < -1) {
                             message(MESS_ERROR,
                                     "%s:%d bad rotation count '%s'\n",
@@ -1307,7 +1320,7 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
                                            &buf, length);
                         if (key == NULL)
                             continue;
-                        newlog->logStart = strtoul(key, &chptr, 0);
+                        newlog->logStart = (int)strtoul(key, &chptr, 0);
                         if (*chptr || newlog->logStart < 0) {
                             message(MESS_ERROR, "%s:%d bad start count '%s'\n",
                                     configFile, lineNum, key);
@@ -1319,7 +1332,7 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
                                            &buf, length);
                         if (key == NULL)
                             continue;
-                        newlog->rotateMinAge = strtoul(key, &chptr, 0);
+                        newlog->rotateMinAge = (int)strtoul(key, &chptr, 0);
                         if (*chptr || newlog->rotateMinAge < 0) {
                             message(MESS_ERROR, "%s:%d bad minimum age '%s'\n",
                                     configFile, lineNum, start);
@@ -1331,7 +1344,7 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
                                            &buf, length);
                         if (key == NULL)
                             continue;
-                        newlog->rotateAge = strtoul(key, &chptr, 0);
+                        newlog->rotateAge = (int)strtoul(key, &chptr, 0);
                         if (*chptr || newlog->rotateAge < 0) {
                             message(MESS_ERROR, "%s:%d bad maximum age '%s'\n",
                                     configFile, lineNum, start);
@@ -1380,6 +1393,8 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
                         scriptDest = &newlog->preremove;
                         state = STATE_LOAD_SCRIPT;
                     } else if (!strcmp(key, "tabooext")) {
+                        char *endtag;
+
                         if (newlog != defConfig) {
                             message(MESS_ERROR,
                                     "%s:%d tabooext may not appear inside "
@@ -1418,7 +1433,7 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
                                 char **tmp = realloc(tabooPatterns, sizeof(*tabooPatterns) *
                                         (tabooCount + 1));
                                 if (tmp == NULL) {
-                                    message(MESS_ERROR, "can not allocate memory\n");
+                                    message_OOM();
                                     RAISE_ERROR();
                                 }
                                 tabooPatterns = tmp;
@@ -1437,6 +1452,8 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
                                 endtag++;
                         }
                     } else if (!strcmp(key, "taboopat")) {
+                        char *endtag;
+
                         if (newlog != defConfig) {
                             message(MESS_ERROR,
                                     "%s:%d taboopat may not appear inside "
@@ -1475,7 +1492,7 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
                             tmp = realloc(tabooPatterns, sizeof(*tabooPatterns) *
                                     (tabooCount + 1));
                             if (tmp == NULL) {
-                                message(MESS_ERROR, "can not allocate memory\n");
+                                message_OOM();
                                 RAISE_ERROR();
                             }
                             tabooPatterns = tmp;
@@ -1493,11 +1510,46 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
                                 endtag++;
                         }
                     } else if (!strcmp(key, "include")) {
+                        int rv;
+
                         free(key);
                         key = isolateValue(configFile, lineNum, "include", &start,
                                            &buf, length);
                         if (key == NULL)
                             continue;
+
+                        if (key[0] == '~' && key[1] == '/') {
+                            /* replace '~' with content of $HOME cause low-level functions
+                             * like stat() do not support the glob ~
+                             */
+                            const char *env_home = secure_getenv("HOME");
+                            char *new_key = NULL;
+
+                            if (!env_home) {
+                                const struct passwd *pwd = getpwuid(getuid());
+                                message(MESS_DEBUG,
+                                        "%s:%d cannot get HOME directory from environment "
+                                        "to replace ~/ in include directive\n",
+                                        configFile, lineNum);
+                                if (!pwd) {
+                                    message(MESS_ERROR, "%s:%d cannot get passwd entry for "
+                                            "running user %u: %s\n",
+                                           configFile, lineNum, getuid(), strerror(errno));
+                                    RAISE_ERROR();
+                                }
+                                env_home = pwd->pw_dir;
+                            }
+
+                            if (asprintf(&new_key, "%s/%s", env_home, key + 2) == -1) {
+                                message_OOM();
+                                RAISE_ERROR();
+                            }
+                            message(MESS_DEBUG, "%s:%d replaced %s with '%s' for include directive\n",
+                                    configFile, lineNum, key, env_home);
+                            free(key);
+                            key = new_key;
+                        }
+
                         message(MESS_DEBUG, "including %s\n", key);
                         if (recursion_depth >= MAX_NESTING) {
                             message(MESS_ERROR, "%s:%d include nesting too deep\n",
@@ -1546,7 +1598,10 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
                                 newlog->addextension);
 
                     } else if (!strcmp(key, "compresscmd")) {
-                        char *compresscmd_base;
+                        char *compresscmd_full;
+                        const char *compresscmd_base;
+                        unsigned i;
+
                         freeLogItem (compress_prog);
 
                         if (!
@@ -1558,11 +1613,14 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
                         message(MESS_DEBUG, "compress_prog is now %s\n",
                                 newlog->compress_prog);
 
-                        compresscmd_base = strdup(basename(newlog->compress_prog));
-                        if (compresscmd_base == NULL) {
-                            message(MESS_ERROR, "can not allocate memory\n");
+                        compresscmd_full = strdup(newlog->compress_prog);
+                        if (compresscmd_full == NULL) {
+                            message_OOM();
                             RAISE_ERROR();
                         }
+
+                        compresscmd_base = basename(compresscmd_full);
+
                         /* we check whether we changed the compress_cmd. In case we use the appropriate extension
                            as listed in compress_cmd_list */
                         for(i = 0; i < compress_cmd_list_size; i++) {
@@ -1570,15 +1628,15 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
                                 freeLogItem (compress_ext);
                                 newlog->compress_ext = strdup(compress_cmd_list[i].ext);
                                 if (newlog->compress_ext == NULL) {
-                                    message(MESS_ERROR, "can not allocate memory\n");
-                                    free(compresscmd_base);
+                                    message_OOM();
+                                    free(compresscmd_full);
                                     RAISE_ERROR();
                                 }
                                 message(MESS_DEBUG, "compress_ext was changed to %s\n", newlog->compress_ext);
                                 break;
                             }
                         }
-                        free(compresscmd_base);
+                        free(compresscmd_full);
                     } else if (!strcmp(key, "uncompresscmd")) {
                         freeLogItem (uncompress_prog);
 
@@ -1643,6 +1701,8 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
                         ) {
                     char *glob_string;
                     size_t glob_count;
+                    int argc, argNum;
+                    const char **argv;
                     in_config = 0;
                     if (newlog != defConfig) {
                         message(MESS_ERROR, "%s:%d unexpected log filename\n",
@@ -1661,7 +1721,7 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
                         newlog->compress_ext = strdup(COMPRESS_EXT);
 
                     if (!newlog->compress_prog || !newlog->uncompress_prog || !newlog->compress_ext) {
-                        message(MESS_ERROR, "can not allocate memory\n");
+                        message_OOM();
                         goto error;
                     }
 
@@ -1694,6 +1754,8 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
                     newlog->numFiles = 0;
                     for (argNum = 0; argNum < argc; argNum++) {
                         char **tmp;
+                        int rc;
+                        glob_t globResult;
 
                         if (globerr_msg) {
                             free(globerr_msg);
@@ -1725,7 +1787,7 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
                                         globResult.
                                         gl_pathc));
                         if (tmp == NULL) {
-                            message(MESS_ERROR, "can not allocate memory\n");
+                            message_OOM();
                             logerror = 1;
                             goto duperror;
                         }
@@ -1733,6 +1795,8 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
                         newlog->files = tmp;
 
                         for (glob_count = 0; glob_count < globResult.gl_pathc; glob_count++) {
+                            struct logInfo *log;
+
                             /* if we glob directories we can get false matches */
                             if (!lstat(globResult.gl_pathv[glob_count], &sb) &&
                                     S_ISDIR(sb.st_mode)) {
@@ -1741,6 +1805,7 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
 
                             for (log = logs.tqh_first; log != NULL;
                                     log = log->list.tqe_next) {
+                                unsigned k;
                                 for (k = 0; k < log->numFiles; k++) {
                                     if (!strcmp(log->files[k],
                                                 globResult.gl_pathv[glob_count])) {
@@ -1757,7 +1822,7 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
                             newlog->files[newlog->numFiles] =
                                 strdup(globResult.gl_pathv[glob_count]);
                             if (newlog->files[newlog->numFiles] == NULL) {
-                                message(MESS_ERROR, "can not allocate memory\n");
+                                message_OOM();
                                 logerror = 1;
                                 goto duperror;
                             }
@@ -1793,14 +1858,16 @@ duperror:
                     }
 
                     if (newlog->oldDir) {
-                        for (i = 0; i < newlog->numFiles; i++) {
+                        unsigned j;
+                        for (j = 0; j < newlog->numFiles; j++) {
                             char *ld;
                             char *dirpath;
                             const char *dirName;
+                            struct stat sb2;
 
-                            dirpath = strdup(newlog->files[i]);
+                            dirpath = strdup(newlog->files[j]);
                             if (dirpath == NULL) {
-                                message(MESS_ERROR, "can not allocate memory\n");
+                                message_OOM();
                                 goto error;
                             }
 
@@ -1827,7 +1894,7 @@ duperror:
                             }
                             ld = malloc(strlen(dirName) + strlen(newlog->oldDir) + 2);
                             if (ld == NULL) {
-                                message(MESS_ERROR, "can not allocate memory\n");
+                                message_OOM();
                                 free(dirpath);
                                 goto error;
                             }
@@ -1879,7 +1946,7 @@ duperror:
                                 message(MESS_ERROR,
                                         "%s:%d olddir %s and log file %s "
                                         "are on different devices\n", configFile,
-                                        lineNum, newlog->oldDir, newlog->files[i]);
+                                        lineNum, newlog->oldDir, newlog->files[j]);
                                 goto error;
                             }
                         }
@@ -1941,13 +2008,13 @@ duperror:
                         state = STATE_SKIP_CONFIG;
                     }
                     else {
-                        endtag = start - 9;
+                        const char *endtag = start - 9;
                         while (*endtag != '\n')
                             endtag--;
                         endtag++;
-                        *scriptDest = strndup(scriptStart, endtag - scriptStart);
+                        *scriptDest = strndup(scriptStart, (size_t)(endtag - scriptStart));
                         if (*scriptDest == NULL) {
-                            message(MESS_ERROR, "can not allocate memory\n");
+                            message_OOM();
                             goto error;
                         }
 
@@ -2012,6 +2079,7 @@ duperror:
             lineNum++;
         }
 
+next_state: ;
     }
 
     if (scriptStart) {
@@ -2023,13 +2091,13 @@ duperror:
 
     free(key);
 
-    munmap(buf, (size_t) length);
+    munmap(buf, length);
     close(fd);
     return logerror;
 error:
     /* free is a NULL-safe operation */
     free(key);
-    munmap(buf, (size_t) length);
+    munmap(buf, length);
     close(fd);
     return 1;
 }
